# coding: utf-8
from __future__ import absolute_import


class ArgumentList(object):

    def __init__(self, owner, base_index=0):
        self.owner = owner
        self.base_index = base_index

    def __getitem__(self, index):
        return self.owner[self.base_index + index]


class Message(list):

    def __init__(self, line, raw=True):
        """General IRC message decoder"""
        if not raw:
            super(Message, self).__init__(line)
        else:
            super(Message, self).__init__()
            # assert isinstance(line, unicode)

            # Append the sender, if not exists
            if line[0] == ':':
                item, line = line.split(' ', 1)
                self.append(decolon(item))
            else:
                self.append(None)

            self += Command(line)

        self.args = ArgumentList(self, 1)

    @property
    def sender(self):
        return self[0]

    @property
    def nick(self):
        return self[0].split('!', 1)[0]

    @property
    def type(self):
        return self[1]


class Command(list):

    def __init__(self, line, raw=True):
        """General IRC message decoder"""
        if not raw:
            super(Command, self).__init__(line)
        else:
            super(Command, self).__init__()

            while ' ' in line:
                # Colon-headed item means it is long argument
                if line[0] == ':':
                    break
                item, line = line.split(' ', 1)
                # Skip zero-lengthed item - split is not good enough for this case
                if len(item) == 0:
                    continue
                self.append(decolon(item))
            # Attach the last item
            self.append(decolon(line))

        self.args = ArgumentList(self, 0)

    @property
    def cmd(self):
        return self[0]


try:
    unicode(0)
except NameError:
    unicode = str  # py3


class Identity(unicode):

    @property
    def nick(self):
        return self.split('!', 1)[0]

    @property
    def username(self):
        return self.split('@', 1)[0].split('!', 1)[1]

    @property
    def host(self):
        return self.split('@', 1)[1]


def decolon(item):
    """Colon has special usage in IRC protocol"""
    if item[0] == ':':
        return item[1:]
    return item



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import threading

from .struct import Message
from .socket import BlockingSocket, NonblockingSocket, ClosedSocket, NoneSocket


class BaseClient(object):
    """
    Common IRC client interface.
    Any client should support:
        self.socket: any ircclient.socket
    """

    def __init__(self, addr, **kwargs):
        blocking = kwargs.get('blocking', True)
        SocketClass = BlockingSocket if blocking else NonblockingSocket
        self.socket = SocketClass(addr)
        self._message_queue = []
        self.options = kwargs
        self.quitting = False

    def runloop_unit(self):
        raise NotImplementedError

    def enqueue(self, msg):
        self._message_queue.append(msg)

    def dispatchable(self):
        return self._message_queue or not self.socket.needs_dispatch()

    def dispatch(self, raw=False):
        """Get a message from socket."""
        if self._message_queue:
            message = self._message_queue.pop(0)
        else:
            message = self.socket.dispatch()
        if message is not None and not raw:
            message = Message(message)
        return message

    def run(self):
        """Thread loop."""
        while self.socket.connected is not False:
            self.runloop_unit()

    def start(self):
        self.thread = threading.Thread()
        self.thread.run = self.run
        self.thread.start()

    def connect(self):
        self.socket.connect()

    def disconnect(self):
        self.socket.disconnect()
        self.socket = ClosedSocket(('', 0))

    def reconnect(self):
        self.socket = NoneSocket()

    def __getattr__(self, key):
        """Borrow commands from command manager."""
        if key.startswith('send_'):
            attr = getattr(self.socket, key)
            return attr
        return self.__getattribute__(key)

    def __call__(self, *args, **kwargs):
        if isinstance(args[0], list):
            self.send_list(args[0])
        else:
            if len(args) == 1:
                if kwargs:
                    line = args[0].format(*args[1:], **kwargs)
                else:
                    line = args[0]
                self.send_line(line)
            else:
                self.send_args(*args)


class DispatchClient(BaseClient):
    """Client implementation based on manual dispatching."""

    def runloop_unit(self):
        """NOTE: blocking"""
        self.socket.recv()


class CallbackClient(BaseClient):
    """Callback-driven IRC client"""

    def __init__(self, addr, callback, options=None):
        self.callback = callback

        BaseClient.__init__(self, addr, options=None)

    def runloop_unit(self):
        """NOTE: blocking"""
        msg = self.dispatch()
        if msg is not None:
            self.callback(self, msg)
        else:
            self.socket.recv()




# Messages
DISCONNECTED = u'DISCONNECTED'
NICK = u'NICK'
USER = u'USER'
PING = u'PING'
PONG = u'PONG'
JOIN = u'JOIN'
PART = u'PART'
QUIT = u'QUIT'
PRIVMSG = u'PRIVMSG'
NOTICE = u'NOTICE'
MODE = u'MODE'
INVITE = u'INVITE'

# Pseudo-messages
# socket level
CREATED = u'CREATED'
CONNECTED = u'CONNECTED'
# event level
LOADED = u'LOADED'

# Message-like
ERROR = u'ERROR'

# From RFC
# sed -n -e "s/ *\([0-9][0-9][0-9]\)  *\([A-Z][A-Z]*_[A-Z][A-Z_]*\)/\2 = u'\1'/p"
# :s/\(.*\) \(.*\)/\2 = u'\1'/
RPL_WELCOME = u'001'
RPL_YOURHOST = u'002'
RPL_CREATED = u'003'
RPL_MYINFO = u'004'
RPL_BOUNCE = u'005'
RPL_USERHOST = u'302'
RPL_ISON = u'303'
RPL_AWAY = u'301'
RPL_UNAWAY = u'305'
RPL_NOWAWAY = u'306'
RPL_WHOISUSER = u'311'
RPL_WHOISSERVER = u'312'
RPL_WHOISOPERATOR = u'313'
RPL_WHOISIDLE = u'317'
RPL_ENDOFWHOIS = u'318'
RPL_WHOISCHANNELS = u'319'
RPL_WHOWASUSER = u'314'
RPL_ENDOFWHOWAS = u'369'
RPL_LISTSTART = u'321'
RPL_LIST = u'322'
RPL_LISTEND = u'323'
RPL_UNIQOPIS = u'325'
RPL_CHANNELMODEIS = u'324'
RPL_NOTOPIC = u'331'
RPL_TOPIC = u'332'
RPL_INVITING = u'341'
RPL_SUMMONING = u'342'
RPL_INVITELIST = u'346'
RPL_ENDOFINVITELIST = u'347'
RPL_EXCEPTLIST = u'348'
RPL_ENDOFEXCEPTLIST = u'349'
RPL_VERSION = u'351'
RPL_WHOREPLY = u'352'
RPL_ENDOFWHO = u'315'
RPL_NAMREPLY = u'353'
RPL_ENDOFNAMES = u'366'
RPL_LINKS = u'364'
RPL_ENDOFLINKS = u'365'
RPL_BANLIST = u'367'
RPL_ENDOFBANLIST = u'368'
RPL_INFO = u'371'
RPL_ENDOFINFO = u'374'
RPL_MOTDSTART = u'375'
RPL_MOTD = u'372'
RPL_ENDOFMOTD = u'376'
RPL_YOUREOPER = u'381'
RPL_REHASHING = u'382'
RPL_YOURESERVICE = u'383'
RPL_TIME = u'391'
RPL_USERSSTART = u'392'
RPL_USERS = u'393'
RPL_ENDOFUSERS = u'394'
RPL_NOUSERS = u'395'
RPL_TRACELINK = u'200'
RPL_TRACECONNECTING = u'201'
RPL_TRACEHANDSHAKE = u'202'
RPL_TRACEUNKNOWN = u'203'
RPL_TRACEOPERATOR = u'204'
RPL_TRACEUSER = u'205'
RPL_TRACESERVER = u'206'
RPL_TRACESERVICE = u'207'
RPL_TRACENEWTYPE = u'208'
RPL_TRACECLASS = u'209'
RPL_TRACERECONNECT = u'210'
RPL_TRACELOG = u'261'
RPL_TRACEEND = u'262'
RPL_STATSLINKINFO = u'211'
RPL_STATSCOMMANDS = u'212'
RPL_ENDOFSTATS = u'219'
RPL_STATSUPTIME = u'242'
RPL_STATSOLINE = u'243'
RPL_UMODEIS = u'221'
RPL_SERVLIST = u'234'
RPL_SERVLISTEND = u'235'
RPL_LUSERCLIENT = u'251'
RPL_LUSEROP = u'252'
RPL_LUSERUNKNOWN = u'253'
RPL_LUSERCHANNELS = u'254'
RPL_LUSERME = u'255'
RPL_ADMINME = u'256'
RPL_ADMINLOC = u'257'
RPL_ADMINLOC = u'258'
RPL_ADMINEMAIL = u'259'
RPL_TRYAGAIN = u'263'
ERR_NOSUCHNICK = u'401'
ERR_NOSUCHSERVER = u'402'
ERR_NOSUCHCHANNEL = u'403'
ERR_CANNOTSENDTOCHAN = u'404'
ERR_TOOMANYCHANNELS = u'405'
ERR_WASNOSUCHNICK = u'406'
ERR_TOOMANYTARGETS = u'407'
ERR_NOSUCHSERVICE = u'408'
ERR_NOORIGIN = u'409'
ERR_NORECIPIENT = u'411'
ERR_NOTEXTTOSEND = u'412'
ERR_NOTOPLEVEL = u'413'
ERR_WILDTOPLEVEL = u'414'
ERR_BADMASK = u'415'
ERR_UNKNOWNCOMMAND = u'421'
ERR_NOMOTD = u'422'
ERR_NOADMININFO = u'423'
ERR_FILEERROR = u'424'
ERR_NONICKNAMEGIVEN = u'431'
ERR_ERRONEUSNICKNAME = u'432'
ERR_NICKNAMEINUSE = u'433'
ERR_NICKCOLLISION = u'436'
ERR_UNAVAILRESOURCE = u'437'
ERR_USERNOTINCHANNEL = u'441'
ERR_NOTONCHANNEL = u'442'
ERR_USERONCHANNEL = u'443'
ERR_NOLOGIN = u'444'
ERR_SUMMONDISABLED = u'445'
ERR_USERSDISABLED = u'446'
ERR_NOTREGISTERED = u'451'
ERR_NEEDMOREPARAMS = u'461'
ERR_ALREADYREGISTRED = u'462'
ERR_NOPERMFORHOST = u'463'
ERR_PASSWDMISMATCH = u'464'
ERR_YOUREBANNEDCREEP = u'465'
ERR_YOUWILLBEBANNED = u'466'
ERR_KEYSET = u'467'
ERR_CHANNELISFULL = u'471'
ERR_UNKNOWNMODE = u'472'
ERR_INVITEONLYCHAN = u'473'
ERR_BANNEDFROMCHAN = u'474'
ERR_BADCHANNELKEY = u'475'
ERR_BADCHANMASK = u'476'
ERR_NOCHANMODES = u'477'
ERR_BANLISTFULL = u'478'
ERR_NOPRIVILEGES = u'481'
ERR_CHANOPRIVSNEEDED = u'482'
ERR_CANTKILLSERVER = u'483'
ERR_RESTRICTED = u'484'
ERR_UNIQOPPRIVSNEEDED = u'485'
ERR_NOOPERHOST = u'491'
ERR_UMODEUNKNOWNFLAG = u'501'
ERR_USERSDONTMATCH = u'502'
RPL_SERVICEINFO = u'231'
RPL_ENDOFSERVICES = u'232'
RPL_SERVICE = u'233'
RPL_NONE = u'300'
RPL_WHOISCHANOP = u'316'
RPL_KILLDONE = u'361'
RPL_CLOSING = u'362'
RPL_CLOSEEND = u'363'
RPL_INFOSTART = u'373'
RPL_MYPORTIS = u'384'
RPL_STATSCLINE = u'213'
RPL_STATSNLINE = u'214'
RPL_STATSILINE = u'215'
RPL_STATSKLINE = u'216'
RPL_STATSQLINE = u'217'
RPL_STATSYLINE = u'218'
RPL_STATSVLINE = u'240'
RPL_STATSLLINE = u'241'
RPL_STATSHLINE = u'244'
RPL_STATSSLINE = u'244'
RPL_STATSPING = u'246'
RPL_STATSBLINE = u'247'
RPL_STATSDLINE = u'250'
ERR_NOSERVICEHOST = u'492'



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import socket
import select

from .const import CONNECTED, DISCONNECTED, CREATED


RAW_LOG = False


class BaseSocket(object):

    def __init__(self, addr, charset='utf-8'):
        """addr is a tuple represents (host, port)"""
        self.addr = addr
        self.charset = charset
        self._buffer = b''
        self.recv_queue = []
        self.send_queue = []
        self.send_callback = None

        self.connected = None
        self.create_socket()

    def create_socket(self):
        """Override to change internal socket."""
        self.socket = socket.socket()

    def send(self, string):
        if self.send_callback:
            self.send_callback(string)
        assert string.endswith('\r\n')
        self._send(string)

    def _send(self, string):
        raise NotImplementedError

    def connect(self):
        self.recv_queue.append(CONNECTED)  # EVERYTHING-IS-RESPONSE!
        self.connected = True
        raise NotImplementedError

    def disconnect(self):
        self.recv_queue.append(DISCONNECTED)  # EVERYTHING-IS-RESPONSE!
        raise NotImplementedError

    def _recv(self):
        raise NotImplementedError

    def send_line(self, line, *arg, **kwargs):
        if len(arg) or len(kwargs):
            line = line.format(*arg, **kwargs)
        # print '--> SEND', line
        self.send(line + '\r\n')

    def send_args(self, *args):
        self.send_list(args)

    def send_list(self, items):
        items = list(items)[:-1] + [':' + items[-1]]
        line = u' '.join(items)
        self.send_line(line)

    def needs_dispatch(self):
        return not self.recv_queue

    def dispatch(self):
        """Dispatch an item from recv_queue."""
        if not self.recv_queue:
            return None
        msg = self.recv_queue.pop(0)
        if RAW_LOG:
            print('>', msg)
        return msg

    def dispatch_all(self):
        msg = True
        while msg:
            msg = self.dispatch()
            if msg:
                yield msg
        raise StopIteration

    def _split_buffer(self):
        """Catch 'ValueError' to check unsplitable"""
        newline, self._buffer = self._buffer.split(b'\r\n', 1)
        try:
            return newline.decode(self.charset)
        except:
            print('Supposed charset is', self.charset, ', but undecodable string found:', newline)
            return newline

    def _enqueue(self, line):
        self.recv_queue.append(line)

    def _enqueue_buffer(self):
        try:
            while True:
                newline = self._split_buffer()
                self._enqueue(newline)
        except ValueError:
            pass

    def recv(self, wait_enqueue=False):
        while True:
            try:
                recv = self._recv()
                if recv is None:
                    if wait_enqueue:
                        continue
                    else:
                        break
                if recv == '':
                    raise socket.error("recv() returned an empty string")
                self._buffer += recv
                self._enqueue_buffer()
            except Exception as e:
                if isinstance(e, socket.error):
                    self.recv_queue.append(DISCONNECTED)
                    self.connected = False
                else:
                    import traceback
                    print('--- Error caught ---')
                    traceback.print_exc()
                    raise e  # debug
                self._enqueue(e)
            if not wait_enqueue or len(self.recv_queue) > 0:
                break


class BlockingSocket(BaseSocket):

    def _send(self, line):
        if RAW_LOG:
            print('<', line,)
        self.socket.send(line.encode(self.charset))

    def connect(self):
        self.socket.connect(self.addr)
        self.recv_queue.append(CONNECTED)  # EVERYTHING-IS-RESPONSE!
        self.connected = True

    def disconnect(self):
        # self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()

    def _recv(self):
        """NOTE: blocking"""
        return self.socket.recv(1024)  # 1kb is enough for irc


class NonblockingSocket(BaseSocket):

    def send(self, line):
        self.send_queue.append(line)

    def connect(self):
        self.socket.connect(self.addr)
        self.recv_queue.append(CONNECTED)  # EVERYTHING-IS-RESPONSE!
        self.connected = True
        self.socket.setblocking(0)

    def disconnect(self):
        # self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()

    def _recv(self):
        if self.send_queue:
            read, write, error = select.select([], [self.socket], [], 0.1)
            if write:
                line = self.send_queue.pop(0)
                if self.send_callback:
                    self.send_callback(line)
                write[0].send(line.encode(self.charset))

        read, write, error = select.select([self.socket], [], [], 0.1)
        if read:
            return read[0].recv(1024)  # 1kb is enough for irc
        else:
            return None


class NoneSocket(BaseSocket):

    def __init__(self, addr, charset='utf-8'):
        self.connected = None

        BaseSocket.__init__(self, addr, charset)

    def dispatch(self):
        return CREATED


class ClosedSocket(BaseSocket):

    def __init__(self, addr, charset='utf-8'):
        self.connected = False

        BaseSocket.__init__(self, addr, charset)

    def dispatch(self):
        return DISCONNECTED
